/* symtab.c   R1 compiler
	P. Chongstitvatana					19 Aug 97
	modify for project A1  				21 Oct 2002
	modify (add field lv) for Som-C  	27 Dec 2003
	add pref for macro					23 Feb 2005
	change to hash for som-v16			7 Nov 2005
*/

#include "compile.h"

int exsym[TABLESIZE], nxsym;		// record export symbol

/* ==========  Interface =============

// search & insert all symtab, return index
int search(char *name)

// insert local symbol at the end of table
int enterLocal(char *name)

// get attribute at index
int getRef(int idx)
int getArg(int idx)
int getLv(int idx)
int getType(int idx)

// set attribute at index
void setRef(int index, int ref)
void setArg(int idx, int arg)
void setLv(int idx, int lv)

// find name of given type and ref, use in listing
char *findName(int type, int ref)

// find name of given index, use in listing
char *getName(int nm)

void dumpSymTab()

 ====================================  */

PRIVATE sym_entry symtab[TABLESIZE+LOCALSIZE];

int lv = 0;

// simply add all char
PRIVATE int hash(char *name){
	int v;
	v = 0;
	while(*name != 0){
		v += ((int)*name)&255;
		name++;
	}
	return v % TABLESIZE;
}
// insert name, return index
PRIVATE int install(char *name){
	int i, key;
	key = hash(name);
	i = key;
	while(1){
		if(symtab[i].name[0] == 0){		// not found, insert
			strcpy(symtab[i].name,name);
			symtab[i].type = tyNEW;
			return i;
		}
		if( eqs(symtab[i].name,name))	// found
			return i;
		i = (i+1) % TABLESIZE;			// linear probing
		if(i == key)
			seterror("symbol table full");
	}
	return 0;		// must not happen
}

/* insert local symbol at the end of table */
int enterLocal(char *name){
	int a;
	lv++;
	if( lv >= LOCALSIZE ) seterror("local symbol table full");
	a = TABLESIZE + lv;
	strcpy(symtab[a].name, name);
	symtab[a].type = tyLOCAL;
	symtab[a].ref = lv;
	symtab[a].arg = 0;
	symtab[a].lv = 0;
	return a;
}

// search all symtab, return idex
// sequential search local at the end first
int search(char *name){
	int i;
	for(i=TABLESIZE+1;i<=TABLESIZE+lv;i++)
		if( eqs(symtab[i].name, name) )
		   return i;
	i = install(name);		// search & insert global
	return i;
}
// symtab access functions
int getRef(int index){ return symtab[index].ref;}
void setRef(int index, int ref){symtab[index].ref = ref;}
int getArg(int idx){ return symtab[idx].arg; }
void setArg(int idx, int arg){ symtab[idx].arg = arg; }
int getLv(int idx){ return symtab[idx].lv; }
void setLv(int idx, int lv){ symtab[idx].lv = lv; }
int getType(int idx){ return symtab[idx].type; }
void setType(int index, int type){symtab[index].type = type;}

// list index of export symbol (tyFUNC, tyGVAR)
// return no. of sym
int getExportSym(int *list){
	int i,k;
	sym_entry *s;
	for(k=0,i=0; i<TABLESIZE; i++){
		s = &symtab[i];
		if( s->type == tyFUNC || s->type == tyGVAR ){
			list[k] = i;
			k++;
		}
	}
	return k;
}
// output symbol table to file
// must getExportSym() first
// update base-ads for static array/string
void dumpSymTab(FILE *FO){
	int i;
	sym_entry *s;
//	k = getExportSym(exsym);
	fprintf(FO,"\n%d\n",nxsym);	// no. of export sym
	for(i=0; i<nxsym; i++){
		s = &symtab[exsym[i]];
//		if(s->arg == VARRAY || s->arg == VSTRING)
//			s->lv = M[s->ref];
		fprintf(FO,"%s %d %d %d %d\n",
			s->name, s->type, s->ref, s->arg, s->lv);
	}
}
// must getExportSym first
char *findName(int type, int ref){
	int i, k;
	for(i=0; i<nxsym; i++){
		k = exsym[i];
		if((symtab[k].type == type) && (symtab[k].ref == ref))
			return symtab[k].name;
	}
	return NULL;
}

char *getName(int nm){return symtab[nm].name;}

void readSymTab(FILE *fi){
	int i, n, idx;
	char name[NAMELEN];
	int type,ref,arg,v;

	fscanf(fi,"%d",&n);
	for(i=0; i<n; i++){
		fscanf(fi,"%s%d%d%d%d",name,&type,&ref,&arg,&v);
		idx = search(name);
		setType(idx,type);
		setRef(idx,ref);
		setArg(idx,arg);
		setLv(idx,v);
		if(type == tyGVAR)
			M[ref] = v;		// set gvar
	}
}

